import numpy as np
import matplotlib.pyplot as plt

from figaro.utils import rejection_sampler

import optparse as op
import importlib

from scipy.stats import norm

from pathlib import Path
import json

def main():
    
    parser = op.OptionParser()
    # Input/output
    parser.add_option("-o", "--output", type = "string", dest = "output", help = "Output folder. Default: working directory", default = '.')
    parser.add_option("-n", "--n_evs", type = "int", dest = "n_evs", help = "Number of events", default = 250)
    parser.add_option("--n_samples", type = "int", dest = "n_samples", help = "Number of samples per event", default = 1000)
    parser.add_option("--pdf", type = "string", dest = "pdf_file", help = "Python module with injected pdf - please name the method 'density'", default = None)
    parser.add_option("--selfunc", type = "string", dest = "selfunc_file", help = "Python module with selection function - please name the method 'selfunc'", default = None)
    parser.add_option("-b", "--bounds", type = "string", dest = "bounds", help = "Minimum and maximum value", default = '-5,5')
    # Plot
    parser.add_option("--symbol", type = "string", dest = "symbol", help = "LaTeX-style quantity symbol, for plotting purposes", default = 'x')
    parser.add_option("--unit", type = "string", dest = "unit", help = "LaTeX-style quantity unit, for plotting purposes", default = None)
    
    (options, args) = parser.parse_args()

    # Paths
    options.output = Path(options.output).resolve()
    options.ev_folder = Path(options.output, 'events').resolve()
    if not options.ev_folder.exists():
        options.ev_folder.mkdir(parents=True)
    
    # PDF
    if options.pdf_file is not None:
        pdf_file_name = Path(options.pdf_file).parts[-1].split('.')[0]
        spec = importlib.util.spec_from_file_location(pdf_file_name, options.pdf_file)
        pdf_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(pdf_module)
        pdf = pdf_module.density
    else:
    	# Default value
        pdf = norm().pdf

    # Selection function
    if options.selfunc_file is not None:
        selfunc_file_name = Path(options.selfunc_file).parts[-1].split('.')[0]
        spec = importlib.util.spec_from_file_location(selfunc_file_name, options.selfunc_file)
        selfunc_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(selfunc_module)
        selfunc = selfunc_module.selection_function
    else:
        selfunc = None
    
    # Bounds
    bounds = [float(b) for b in options.bounds.split(',')]

    # Draw true values
    true_vals = rejection_sampler(options.n_evs, pdf, bounds, selfunc = selfunc)
    # Draw posterior widths
    widths = np.random.uniform(2,4, size = options.n_evs)
    
    true_vals_dict = {}
    all_samples    = []
    for i in range(options.n_evs):
    
        # Store true value
        true_vals_dict[str(i+1)] = true_vals[i]
        
        # Draw samples
        mean_val = np.random.normal(true_vals[i], widths[i], 1)
        samples  = np.random.normal(mean_val, widths[i], options.n_samples)
        all_samples.extend(samples)
        
        # Save samples
        np.savetxt(Path(options.ev_folder,'{0}.txt'.format(i+1)), samples)
    
    #Save true values as JSON file
    with open(Path(options.output, 'true_vals.json'), 'w', encoding='utf-8') as f:
        json.dump(true_vals_dict, f, ensure_ascii=False)
    # Save true values as txt file as well
    np.savetxt(Path(options.output, 'truths.txt'), true_vals)
    
    x  = np.linspace(bounds[0],bounds[1],1000)
    dx = x[1]-x[0]
    # Make plot
    plt.hist(all_samples, bins = int(np.sqrt(len(all_samples))), density = True, histtype = 'step', label = 'Samples (stacked)', color = 'forestgreen', alpha = 0.5)
    plt.hist(true_vals, bins = int(np.sqrt(len(true_vals))), density = True, histtype = 'step', label = 'True values', color = 'steelblue')
    plt.plot(x, pdf(x), c = 'r', lw = 0.7, label = 'Injected')
    
    if selfunc is not None:
        filtered = density(x)*selection_function(x)
        plt.plot(x, filtered/np.sum(filtered*dx), c = 'k', lw = 0.7, label = 'After selection effects')
    
    if options.unit is not None:
        plt.xlabel('${0}\ [{1}]$'.format(options.symbol, options.unit))
    else:
        plt.xlabel('${0}$'.format(options.symbol))
    plt.ylabel('$p({0})$'.format(options.symbol))
    plt.grid(alpha = 0.5)
    plt.legend(loc=0, frameon=False)
    plt.savefig(Path(options.output, 'truths.pdf'), bbox_inches = 'tight')

if __name__ == '__main__':
    main()



import numpy as np
from scipy.special import erfinv, erf

log2PI = np.log(2.0*np.pi)

"""
Scipy's implementation of ERF: https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.erf.html
See notes there for cumulative of the unit normal distribution.
"""

def transform_to_probit(x, bounds):
    '''
    Coordinate change into probit space.
    cdf_normal is the cumulative distribution function of the unit normal distribution.
    WARNING: returns NAN if x is not in [xmin, xmax].
    
    t(x) = cdf^-1_normal((x-x_min)/(x_max - x_min))

    
    Arguments:
        :np.ndarray x:      sample(s) to transform (2d array)
        :np.ndarray bounds: limits for each dimension (2d array, [[xmin, xmax], [ymin, ymax]...])
        
    Returns:
        :np.ndarray: sample(s)
    '''
    dbounds = bounds[:,1]-bounds[:,0]
    sigma   = dbounds*0.34
    cdf = (x - bounds[:,0])/dbounds
    o = np.sqrt(2.0)*erfinv(2*cdf-1)*sigma
    return o

def transform_from_probit(x, bounds):
    '''
    Coordinate change from probit to natural space.
    cdf_normal is the cumulative distribution function of the unit normal distribution.
    
    x(t) = xmin + (xmax-xmin)*cdf_normal(t|0,1)
    
    Arguments:
        :np.ndarray x:      sample(s) to antitransform (2d array)
        :np.ndarray bounds: limits for each dimension (2d array, [[xmin, xmax], [ymin, ymax]...])
        
    Returns:
        :np.ndarray: sample(s)
    '''
    dbounds = bounds[:,1]-bounds[:,0]
    sigma   = dbounds*0.34
    cdf = 0.5*(1.0+erf(x/(np.sqrt(2.0)*sigma)))
    o = bounds[:,0]+dbounds*cdf
    return o

def probit_log_jacobian(x, bounds):
    dbounds = bounds[:,1]-bounds[:,0]
    sigma   = dbounds*0.34
    return -0.5*(x/sigma)**2-0.5*(log2PI + np.log(sigma))+np.log(dbounds)

def probit_logJ(x, bounds, flag = True):
    if not flag:
        return np.zeros(len(x))
    dbounds = bounds[:,1]-bounds[:,0]
    sigma   = dbounds*0.34
    res     = np.sum(-0.5*(x/sigma)**2-0.5*(log2PI + np.log(sigma))+np.log(dbounds), axis = -1)
    return res



import numpy as np
from numba import jit, njit, prange

LOG2PI = np.log(2*np.pi)

#-----------#
# Functions #
#-----------#

@njit
def inv_jit(M):
  return np.linalg.inv(M)

@njit
def logdet_jit(M):
    return np.log(np.linalg.det(M))

@jit
def logsumexp_jit(a, b):
    a_max = np.max(a)
    tmp = b * np.exp(a - a_max)
    return np.log(np.sum(tmp)) + a_max

@njit
def scalar_product(v, M, n):
    """
    Scalar product: v*M*v^T
    
    Arguments:
        :np.ndarray v: array
        :np.ndarray M: matrix
        :int n:        len(v)
    
    Returns:
        :double: v*M*v^T
    """
    res = 0.
    for i in prange(n):
        for j in prange(n):
            res = res + M[i,j]*v[i]*v[j]
    return res

@jit
def log_norm_1d(x, m, s):
    """
    1D Normal logpdf
    
    Arguments:
        :double x: value
        :double m: mean
        :double s: var
    
    Returns:
        Normal(m,s).logpdf(x)
    """
    return -(x-m)**2/(2*s) - 0.5*np.log(2*np.pi*s)

@jit
def log_norm(x, mu, cov):
    """
    Multivariate Normal logpdf
    
    Arguments:
        :np.ndarray x:   value
        :np.ndarray mu:  mean vector
        :np.ndarray cov: covariance matrix
    
    Returns:
        :double: MultivariateNormal(m,s).logpdf(x)
    """
    inv_cov  = inv_jit(cov)
    exponent = -0.5*scalar_product(x-mu, inv_cov, len(mu))
    lognorm  = 0.5*len(mu)*LOG2PI+0.5*logdet_jit(cov)
    return -lognorm+exponent

@jit
def log_norm_int(x, mu, cov_1, inv_cov_1, cov_2):
    """
    Multivariate Normal logpdf (tailored to this problem!)
    See https://arxiv.org/pdf/1811.04751v1.pdf
    
    Arguments:
        :np.ndarray x:         value
        :np.ndarray mu:        mean vector
        :np.ndarray cov_1:     1st covariance matrix
        :np.ndarray inv_cov_1: inverse of 1st covariance matrix
        :np.ndarray cov_2:     2nd covariance matrix
    
    Returns:
        :double: MultivariateNormal(m,s).logpdf(x)
    """
    inv_cov_2  = inv_jit(cov_2)
    inv_cov    = inv_cov_1@inv_jit(inv_cov_1+inv_cov_2)@inv_cov_2
    exponent   = -0.5*scalar_product(x-mu, inv_cov, len(mu))
    lognorm    = 0.5*len(mu)*LOG2PI+0.5*(logdet_jit(cov_1) + logdet_jit(cov_2) + logdet_jit(inv_cov_1+inv_cov_2))
    return -lognorm+exponent

#------------#
# 1D methods #
#------------#

@jit
def eval_mix_1d(mu, sigma, means, covs):
    """
    Computes N(mu_k| mu, (sigma_k^2+sigma^2) for all the components of a mixture (for predictive likelihood, 1D).
    
    Arguments:
        :np.ndarray mu:    temptative mean of the parent mixture component
        :np.ndarray sigma: temptative variance of the parent mixture component
        :np.ndarray means: means of the event mixture components
        :np.ndarray vars:  variances of the event mixture components
    
    Returns:
        :np.ndarray: probability for each event mixture components
    """
    return np.array([log_norm_1d(means[i,0], mu, sigma+covs[i,0,0]) for i in prange(len(means))])

@jit
def evaluate_mixture_MC_draws_1d(mu, sigma, means, vars, w):
    """
    Computes N(mu_k| mu, (sigma_k^2+sigma^2) for a set of MC draws for mu and sigma.
    
    Arguments:
        :np.ndarray mu:    MC draws for the mean of the parent mixture component
        :np.ndarray sigma: MC draws for the variance of the parent mixture component
        :np.ndarray means: means of the event mixture components
        :np.ndarray vars:  variances of the event mixture components
        :np.ndarray w:     component weights
    
    Returns:
        :np.ndarray: probability for each MC draw
    """
    logP = np.zeros(len(mu), dtype = np.float64)
    for i in prange(len(mu)):
        logP[i] = logsumexp_jit(eval_mix_1d(mu[i], sigma[i], means, vars), b = w)
    return logP

#------------#
# ND methods #
#------------#

@jit
def eval_mix(mu, sigma, means, covs):
    """
    Computes N(mu_k| mu, (sigma_k^2+sigma^2) for all the components of a mixture (for predictive likelihood, ND).
    
    Arguments:
        :np.ndarray mu:    temptative mean of the parent mixture component
        :np.ndarray sigma: temptative covariance matrix of the parent mixture component
        :np.ndarray means: means of the event mixture components
        :np.ndarray covs:  covariance matrices of the event mixture components
    
    Returns:
        :np.ndarray: probability for each event mixture components
    """
    inv_sigma = inv_jit(sigma)
    return np.array([log_norm_int(means[i], mu, sigma, inv_sigma, covs[i]) for i in prange(len(means))])

@jit
def evaluate_mixture_MC_draws(mu, sigma, means, covs, w):
    """
    Computes N(mu_k| mu, (sigma_k^2+sigma^2) for a set of MC draws for mu and sigma.
    
    Arguments:
        :np.ndarray mu:    MC draws for the mean vector of the parent mixture component
        :np.ndarray sigma: MC draws for the covariance matrix of the parent mixture component
        :np.ndarray means: means of the event mixture components
        :np.ndarray covs:  covariance matrices of the event mixture components
        :np.ndarray w:     component weights
    
    Returns:
        :np.ndarray: probability for each MC draw
    """
    logP = np.zeros(len(mu), dtype = np.float64)
    for i in prange(len(mu)):
        logP[i] = logsumexp_jit(eval_mix(mu[i], sigma[i], means, covs), b = w)
    return logP



import numpy as np
cimport numpy as np
cdef extern from "lal/LALCosmologyCalculator.h" nogil:
    # cosmological parameters structure
    ctypedef struct LALCosmologicalParameters:
        double h;
        double om;
        double ol;
        double ok;
        double w0;
        double w1;
        double w2;

    cdef double XLALLuminosityDistance(
            LALCosmologicalParameters *omega,
            double z)

    cdef double XLALAngularDistance(
            LALCosmologicalParameters *omega,
            double z)

    cdef double XLALComovingLOSDistance(
            LALCosmologicalParameters *omega,
            double z)

    cdef double XLALComovingTransverseDistance(
            LALCosmologicalParameters *omega,
            double z)

    cdef double XLALHubbleDistance(
            LALCosmologicalParameters *omega
            )

    cdef double XLALHubbleParameter(double z,
            void *omega
            )

    cdef double XLALIntegrateHubbleParameter(LALCosmologicalParameters *omega, double z)

    cdef double XLALUniformComovingVolumeDistribution(LALCosmologicalParameters *omega, double z, double zmax)

    cdef double XLALUniformComovingVolumeDensity(double z, void *omega)

    cdef double XLALIntegrateComovingVolumeDensity(LALCosmologicalParameters *omega, double z)

    cdef double XLALIntegrateComovingVolume(LALCosmologicalParameters *omega, double z)

    cdef double XLALComovingVolumeElement(double z, void *omega)

    cdef double XLALComovingVolume(LALCosmologicalParameters *omega, double z)

    cdef LALCosmologicalParameters *XLALCreateCosmologicalParameters(double h, double om, double ol, double w0, double w1, double w2)

    cdef void XLALDestroyCosmologicalParameters(LALCosmologicalParameters *omega)

    cdef double XLALGetHubbleConstant(LALCosmologicalParameters *omega)

    cdef double XLALGetOmegaMatter(LALCosmologicalParameters *omega)

    cdef double XLALGetOmegaLambda(LALCosmologicalParameters *omega)

    cdef double XLALGetOmegaK(LALCosmologicalParameters *omega)

    cdef double XLALGetW0(LALCosmologicalParameters *omega)

    cdef double XLALGetW1(LALCosmologicalParameters *omega)

    cdef double XLALGetW2(LALCosmologicalParameters *omega)
    
cdef class CosmologicalParameters:
    cdef LALCosmologicalParameters* _LALCosmologicalParameters
    cdef public double h
    cdef public double om
    cdef public double ol
    cdef public double w0
    cdef public double w1
    cdef public double _HubbleParameter(self,double z) nogil
    cdef public double _LuminosityDistance_double(self, double z) nogil
    cdef public np.ndarray[double, ndim=1, mode="c"] _LuminosityDistance(self, np.ndarray[double, ndim=1, mode="c"] z)
    cdef public double _HubbleDistance(self) nogil
    cdef public double _IntegrateComovingVolumeDensity(self, double zmax) nogil
    cdef public double _IntegrateComovingVolume(self, double zmax) nogil
    cdef public double _UniformComovingVolumeDensity_double(self, double z) nogil
    cdef public np.ndarray[double, ndim=1, mode="c"] _UniformComovingVolumeDensity(self, np.ndarray[double, ndim=1, mode="c"] z)
    cdef public double _UniformComovingVolumeDistribution(self, double z, double zmax) nogil
    cdef public double _ComovingVolumeElement_double(self,double z) nogil
    cdef public np.ndarray[double, ndim=1, mode="c"] _ComovingVolumeElement(self, np.ndarray[double, ndim=1, mode="c"] z)
    cdef public double _ComovingVolume_double(self, double z) nogil
    cdef public np.ndarray[double, ndim=1, mode="c"] _ComovingVolume(self, np.ndarray[double, ndim=1, mode="c"] z)
    cdef void _DestroyCosmologicalParameters(self) nogil



import numpy as np
from numba import njit
from figaro.exceptions import FIGAROException
from figaro.likelihood import log_norm, inv_jit
from figaro.decorators import probit

@njit
def _cond_mean_cov(vals, mu1, mu2, s11, s22, s12):
    """
    Compute mean and covariance for conditional Normal distribution.
    See https://stats.stackexchange.com/questions/348941/general-conditional-distributions-for-multivariate-gaussian-mixtures
    
    Arguments:
        :np.ndarray vals: values to condition on
        :np.ndarray mu1:  mean subvector (NOT conditioned)
        :np.ndarray mu2:  mean subvector (conditioned)
        :np.ndarray s11:  covariance submatrix (NOT conditioned)
        :np.ndarray s22:  covariance submatrix (conditioned)
        :np.ndarray s12:  off-diagonal
    
    Returns:
        :np.ndarray: mean
        :np.ndarray: covariance
    """
    s22_inv = inv_jit(s22)
    mean = mu1 + s12@(s22_inv@(vals - mu2))
    cov  = s11 - s12@(s22_inv@s12.T)
    return mean, cov

def _marginalise(mix, axis = -1):
    """
    Marginalise out one or more dimensions from a FIGARO mixture.
    
    Arguments:
        :figaro.mixture.mixture draws: mixture
        :int or list of int axis:      axis to marginalise on
    
    Returns:
        :figaro.mixture.mixture: the marginalised mixture
    """
    # Circular import
    from figaro.mixture import mixture
    ax     = np.atleast_1d(axis)
    dim    = mix.dim - len(ax)
    if dim < 1:
        raise FIGAROException("Cannot marginalise out all dimensions")
    means  = np.delete(mix.means, ax, axis = -1)
    covs   = np.delete(np.delete(mix.covs, ax, axis = -1), ax, axis = -2)
    bounds = np.delete(mix.bounds, ax, axis = 0)
    
    return mixture(means, covs, mix.w, bounds, dim, mix.n_cl, mix.n_pts, probit = mix.probit)

def marginalise(draws, axis = -1):
    """
    Marginalise out one or more dimensions from a FIGARO draw.
    
    Arguments:
        :figaro.mixture.mixture draws: mixture(s)
        :int or list of int axis:      axis to marginalise on
    
    Returns:
        :figaro.mixture.mixture: the marginalised mixture(s)
    """
    if axis == []:
        return draws
    if np.iterable(draws):
        return np.array([_marginalise(d, axis) for d in draws])
    else:
        return _marginalise(draws, axis)

@probit
def _condition(mix, vals, dims, norm = True):
    """
    Probability density conditioned on specific values of a subset of parameters.
    See:
     * https://stats.stackexchange.com/questions/348941/general-conditional-distributions-for-multivariate-gaussian-mixtures
     * https://stats.stackexchange.com/questions/30588/deriving-the-conditional-distributions-of-a-multivariate-normal-distribution
    
    Arguments:
        :figaro.mixture.mixture mix: mixture
        :iterable vals:              value(s) to condition on
        :int or list of int dims:    dimension(s) associated with given vals (starting from 0)
        :bool norm:                  normalize the distribution
    
    Returns:
        :figaro.mixture.mixture: the conditioned mixture(s)
    """
    # Circular import
    from figaro.mixture import mixture
    ax   = np.atleast_1d(dims)
    vals = vals[ax]
    dim  = mix.dim - len(ax)
    idx  = np.array([i in ax for i in range(mix.dim)])
    if dim < 1:
        raise FIGAROException("Cannot condition on all dimensions")
    bounds    = np.delete(mix.bounds, ax, axis = 0)
    means     = np.zeros(shape = (mix.n_cl, dim))
    covs      = np.zeros(shape = (mix.n_cl, dim, dim))
    weights   = np.zeros(shape = mix.n_cl)
    for i, (mu, cov) in enumerate(zip(mix.means, mix.covs)):
        # Subvectors and submatrices
        mu1 = mu[~idx]
        mu2 = mu[idx]
        s11 = cov[~idx,:][:,~idx]
        s12 = cov[idx,:][:,~idx]
        s22 = cov[idx,:][:,idx]
        # Parameters
        means[i], covs[i] = _cond_mean_cov(vals, mu1, mu2, s11, s22, s12)
        weights[i] = np.exp(log_norm(vals, mu2, s22))
    # Weights
    weights = mix.w*weights
    if norm:
        weights /= _marginalise(mix, axis = np.arange(mix.dim)[~idx]).pdf(vals)
    return mixture(means, covs, weights, bounds, dim, mix.n_cl, mix.n_pts, probit = mix.probit)

def condition(draws, vals, dims, norm = True):
    """
    Probability density conditioned on specific values of a subset of parameters.
    
    Arguments:
        :figaro.mixture.mixture draws: mixture(s)
        :iterable vals:                value(s) to condition on
        :int or list of int dims:      dimension(s) associated with given vals (starting from 0)
        :bool norm:                    normalize the distribution
    
    Returns:
        :figaro.mixture.mixture: the conditioned mixture(s)
    """
    if np.iterable(draws):
        v       = np.mean(draws[0].bounds, axis = -1)
        v[dims] = vals
        return np.array([_condition(d, v, dims, norm) for d in draws])
    else:
        v       = np.mean(draws.bounds, axis = -1)
        v[dims] = vals
        return _condition(draws, v, dims, norm)



from matplotlib import rcParams
from distutils.spawn import find_executable

tex_flag = False
if find_executable('latex'):
    rcParams["text.usetex"] = True
    tex_flag = True
rcParams["xtick.labelsize"] = 14
rcParams["ytick.labelsize"] = 14
rcParams["xtick.direction"] = "in"
rcParams["ytick.direction"] = "in"
rcParams["legend.fontsize"] = 12
rcParams["legend.frameon"]  = False
rcParams["axes.labelsize"]  = 16
rcParams["axes.grid"]       = True
rcParams["grid.alpha"]      = 0.6
rcParams["grid.linestyle"]  = "dotted"
rcParams["lines.linewidth"] = 0.7
rcParams["contour.negative_linestyle"] = "solid"



import numpy as np
cimport numpy as np
cimport cython

cdef class CosmologicalParameters:

    def __cinit__(self, double h, double om, double ol, double w0, double w1):
        self.h = h
        self.om = om
        self.ol = ol
        self.w0 = w0
        self.w1 = w1
        self._LALCosmologicalParameters = XLALCreateCosmologicalParameters(self.h,self.om,self.ol,self.w0,self.w1,0.0)
    def __reduce__(self):
        return (CosmologicalParameters, (self.h, self.om, self.ol, self.w0, self.w1))

    cdef double _HubbleParameter(self, double z) nogil:
        return XLALHubbleParameter(z, self._LALCosmologicalParameters)

    cdef double _LuminosityDistance_double(self, double z) nogil:
        return XLALLuminosityDistance(self._LALCosmologicalParameters,z)

    cdef np.ndarray[double, ndim=1, mode="c"] _LuminosityDistance(self, np.ndarray[double, ndim=1, mode="c"] z):
        cdef unsigned int i, n = z.shape[0]
        cdef np.ndarray[double, ndim=1, mode="c"] DL = np.zeros(n)
        cdef double[:] DL_view = DL
        for i in range(n):
            DL_view[i] = self._LuminosityDistance_double(z[i])
        return DL

    cdef double _HubbleDistance(self) nogil:
        return XLALHubbleDistance(self._LALCosmologicalParameters)

    cdef double _IntegrateComovingVolumeDensity(self, double zmax) nogil:
        return XLALIntegrateComovingVolumeDensity(self._LALCosmologicalParameters,zmax)

    cdef double _IntegrateComovingVolume(self, double zmax) nogil:
        return XLALIntegrateComovingVolume(self._LALCosmologicalParameters,zmax)

    cdef double _UniformComovingVolumeDensity_double(self, double z) nogil:
        return XLALUniformComovingVolumeDensity(z, self._LALCosmologicalParameters)
    
    cdef np.ndarray[double, ndim=1, mode="c"] _UniformComovingVolumeDensity(self, np.ndarray[double, ndim=1, mode="c"] z):
        cdef unsigned int i, n = z.shape[0]
        cdef np.ndarray[double, ndim=1, mode="c"] p_z = np.zeros(n)
        cdef double[:] p_z_view = p_z
        for i in range(n):
            p_z_view[i] = self._UniformComovingVolumeDensity_double(z[i])
        return p_z

    cdef double _UniformComovingVolumeDistribution(self, double z, double zmax) nogil:
        return XLALUniformComovingVolumeDistribution(self._LALCosmologicalParameters, z, zmax)

    cdef double _ComovingVolumeElement_double(self,double z) nogil:
        return XLALComovingVolumeElement(z, self._LALCosmologicalParameters)
        
    cdef np.ndarray[double, ndim=1, mode="c"] _ComovingVolumeElement(self, np.ndarray[double, ndim=1, mode="c"] z):
        cdef unsigned int i, n = z.shape[0]
        cdef np.ndarray[double, ndim=1, mode="c"] dVdz = np.zeros(n)
        cdef double[:] dVdz_view = dVdz
        for i in range(n):
            dVdz_view[i] = self._ComovingVolumeElement_double(z[i])
        return dVdz

    cdef double _ComovingVolume_double(self, double z) nogil:
        return XLALComovingVolume(self._LALCosmologicalParameters, z)

    cdef np.ndarray[double, ndim=1, mode="c"] _ComovingVolume(self, np.ndarray[double, ndim=1, mode="c"] z):
        cdef unsigned int i, n = z.shape[0]
        cdef np.ndarray[double, ndim=1, mode="c"] V = np.zeros(n)
        cdef double[:] V_view = V
        for i in range(n):
            V_view[i] = self._ComovingVolume_double(z[i])
        return V

    cdef void _DestroyCosmologicalParameters(self) nogil:
        XLALDestroyCosmologicalParameters(self._LALCosmologicalParameters)
        return

    def HubbleParameter(self, double z):
        return self._HubbleParameter(z)

    def LuminosityDistance_double(self, double z):
        return self._LuminosityDistance_double(z)
    
    def LuminosityDistance(self, np.ndarray[double, ndim=1, mode="c"] z):
        return self._LuminosityDistance(z)

    def HubbleDistance(self):
        return self._HubbleDistance()

    def IntegrateComovingVolumeDensity(self, double zmax):
        return self._IntegrateComovingVolumeDensity(zmax)

    def IntegrateComovingVolume(self, double zmax):
        return self._IntegrateComovingVolume(zmax)

    def UniformComovingVolumeDensity_double(self, double z):
        return self._UniformComovingVolumeDensity_double(z)

    def UniformComovingVolumeDensity(self, np.ndarray[double, ndim=1, mode="c"] z):
        return self._UniformComovingVolumeDensity(z)

    def UniformComovingVolumeDistribution(self, double z, double zmax):
        return self._UniformComovingVolumeDistribution(z, zmax)

    def ComovingVolumeElement_double(self, double z):
        return self._ComovingVolumeElement_double(z)
        
    def ComovingVolumeElement(self, np.ndarray[double, ndim=1, mode="c"] z):
        return self._ComovingVolumeElement(z)

    def ComovingVolume_double(self, double z):
        return self._ComovingVolume_double(z)

    def ComovingVolume(self, np.ndarray[double, ndim=1, mode="c"] z):
        return self._ComovingVolume(z)

    def DestroyCosmologicalParameters(self):
        self._DestroyCosmologicalParameters()
        return



import numpy as np
import warnings

from pathlib import Path
from scipy.stats import beta
from corner import corner

import matplotlib.pyplot as plt
from matplotlib import axes
from matplotlib.projections import projection_registry
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
from matplotlib import rcParams

from figaro import plot_settings
from figaro.marginal import marginalise
from figaro.credible_regions import ConfidenceArea
from figaro.utils import recursive_grid

# Telling python to ignore empty legend warning from matplotlib
warnings.filterwarnings("ignore", message = "No artists with labels found to put in legend.  Note that artists whose label start with an underscore are ignored when legend() is called with no argument.")

class PPPlot(axes.Axes):
    """
    Construct a probability-probability (P-P) plot.

    Derived from https://lscsoft.docs.ligo.org/ligo.skymap/_modules/ligo/skymap/plot/pp.html#PPPlot
    This class avoids installing the whole ligo.skymap.plot package.
    """

    name = 'pp_plot'

    def __init__(self, *args, **kwargs):
        # Call parent constructor
        super().__init__(*args, **kwargs)

        # Square axes, limits from 0 to 1
        self.set_aspect(1.0)
        self.set_xlim(0.0, 1.0)
        self.set_ylim(0.0, 1.0)

    def add_diagonal(self, *args, **kwargs):
        """
        Add a diagonal line to the plot, running from (0, 0) to (1, 1).

        Other parameters
        ----------------
        kwargs :
            optional extra arguments to `matplotlib.axes.Axes.plot`
        """
        # Make copy of kwargs to pass to plot()
        kwargs = dict(kwargs)
        kwargs.setdefault('color', 'black')
        kwargs.setdefault('linestyle', 'dashed')
        kwargs.setdefault('linewidth', 0.5)

        # Plot diagonal line
        return self.plot([0, 1], [0, 1], *args, **kwargs)

    def add_confidence_band(self, nsamples, cl=0.9, **kwargs):
        """
        Add a target confidence band.

        Parameters
        ----------
        nsamples : int
            Number of P-values
        cl : float, default: 0.9
            Confidence level

        Other parameters
        ----------------
        **kwargs :
            optional extra arguments to `matplotlib.axes.Axes.fill_betweenx`
        """
        n = nsamples
        k = np.arange(0, n + 1)
        p = k / n
        ci_lo, ci_hi = beta.interval(cl, k + 1, n - k + 1)

        # Make copy of kwargs to pass to fill_betweenx()
        kwargs = dict(kwargs)
        kwargs.setdefault('color', 'ghostwhite')
        kwargs.setdefault('edgecolor', 'gray')
        kwargs.setdefault('linewidth', 0.5)
        kwargs.setdefault('alpha', 0.5)
        fontsize = kwargs.pop('fontsize', 'x-small')

        return self.fill_betweenx(p, ci_lo, ci_hi, **kwargs, label = '${0}\%\ CR$'.format(int(cl*100)))

    @classmethod
    def _as_mpl_axes(cls):
        """
        Support placement in figure using the `projection` keyword argument.
        See http://matplotlib.org/devel/add_new_projection.html.
        """
        return cls, {}
        
projection_registry.register(PPPlot)

def plot_median_cr(draws, injected = None, samples = None, selfunc = None, bounds = None, out_folder = '.', name = 'density', n_pts = 1000, label = None, unit = None, hierarchical = False, show = False, save = True, subfolder = False, true_value = None, true_value_label = '\mathrm{True\ value}', injected_label = '\mathrm{Simulated}', median_label = None):
    """
    Plot the recovered 1D distribution along with the injected distribution and samples from the true distribution (both if available).
    
    Arguments:
        :iterable draws:                  container for mixture instances
        :callable or np.ndarray injected: injected distribution (if available)
        :np.ndarray samples:              samples from the true distribution (if available)
        :callable or np.ndarray selfunc:  selection function (if available)
        :iterable bounds:                 bounds for the recovered distribution. If None, bounds from mixture instances are used.
        :str or Path out_folder:          output folder
        :str name:                        name to be given to outputs
        :int n_pts:                       number of points for linspace
        :str label:                       LaTeX-style quantity label, for plotting purposes
        :str unit:                        LaTeX-style quantity unit, for plotting purposes
        :bool hierarchical:               hierarchical inference, for plotting purposes
        :bool save:                       whether to save the plots or not
        :bool show:                       whether to show the plots during the run or not
        :bool subfolder:                  whether to save the plots in different subfolders (for multiple events)
        :float true_value:                true value to infer
        :str true_value_label:            label to assign to the true value marker
        :str injected_label:              label to assign to the injected distribution
        :str median_label:                label to assign to the reconstruction
    """
    if median_label is None:
        if hierarchical:
            median_label = '\mathrm{(H)DPGMM}'
        else:
            median_label = '\mathrm{DPGMM}'
    
    all_bounds = np.atleast_2d([d.bounds[0] for d in draws])
    x_min = np.max(all_bounds[:,0])
    x_max = np.min(all_bounds[:,1])
    
    probit = np.array([d.probit for d in draws]).any()
    
    if bounds is not None:
        if not bounds[0] >= x_min and probit:
            warnings.warn("The provided lower bound is invalid for at least one draw. {0} will be used instead.".format(x_min))
        else:
            x_min = bounds[0]
        if not bounds[1] <= x_max and probit:
            warnings.warn("The provided upper bound is invalid for at least one draw. {0} will be used instead.".format(x_max))
        else:
            x_max = bounds[1]

    fig, ax = plt.subplots()
    # If samples are available, use them as bounds
    if samples is not None:
        ax.hist(samples, bins = int(np.sqrt(len(samples))), histtype = 'step', density = True, label = '$\mathrm{Samples}$', log = True)
        x_min_l, x_max_l = ax.get_xlim()
        x_min = np.max((x_min, x_min_l))
        x_max = np.min((x_max, x_max_l))
        
    x    = np.linspace(x_min, x_max, n_pts+2)[1:-1]
    dx   = x[1]-x[0]
    
    probs = np.array([d.pdf(x) for d in draws])
    
    percentiles = [50, 5, 16, 84, 95]
    p = {}
    for perc in percentiles:
        p[perc] = np.percentile(probs, perc, axis = 0)
    norm = p[50].sum()*dx
    for perc in percentiles:
        p[perc] = p[perc]/norm

    # Samples (if available)
    if samples is not None:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
    else:
        ax.set_yscale('log')
    
    # CR
    ax.fill_between(x, p[95], p[5], color = 'mediumturquoise', alpha = 0.25)
    ax.fill_between(x, p[84], p[16], color = 'darkturquoise', alpha = 0.25)
    # Selection function (if available)
    if selfunc is not None:
        if callable(selfunc):
            f_x = selfunc(x)
        else:
            f_x = selfunc
    # Injection (if available)
    if injected is not None:
        if callable(injected):
            p_x = injected(x)
        else:
            p_x = injected
        if injected_label is not None:
            injected_label = '$'+injected_label+'$'
        ax.plot(x, p_x, lw = 0.5, color = 'red', label = injected_label)
        if selfunc is not None:
            filtered_p_x = p_x*f_x
            ax.plot(x, filtered_p_x/np.sum(filtered_p_x*dx), lw = 0.5, color = 'k', label = '$\mathrm{Selection\ effects}$')
        
    # Median
    if true_value is not None:
        if true_value_label is not None:
            true_value_label = '$'+true_value_label+'$'
        ax.axvline(true_value, ls = '--', color = 'r', lw = 0.5, label = true_value_label)
    ax.plot(x, p[50], lw = 0.7, color = 'steelblue', label = '${0}$'.format(median_label))
    if label is None:
        label = 'x'
    if unit is None or unit == '':
        ax.set_xlabel('${0}$'.format(label))
    else:
        ax.set_xlabel('${0}\ [{1}]$'.format(label, unit))
    ax.set_ylabel('$p({0})$'.format(label))
    if samples is not None:
        ax.set_xlim(xlim)
    ax.set_ylim(bottom = 1e-5, top = np.max(p[95])*1.1)
    ax.legend(loc = 0)
    
    fig.align_labels()
    
    if save:
        if subfolder:
            plot_folder = Path(out_folder, 'density')
            if not plot_folder.exists():
                try:
                    plot_folder.mkdir()
                except FileExistsError:
                    # Avoids issue with parallelisation
                    pass
            log_folder = Path(out_folder, 'log_density')
            if not log_folder.exists():
                try:
                    log_folder.mkdir()
                except FileExistsError:
                    pass
            txt_folder = Path(out_folder, 'txt')
            if not txt_folder.exists():
                try:
                    txt_folder.mkdir()
                except FileExistsError:
                    pass
        else:
            plot_folder = out_folder
            log_folder  = out_folder
            txt_folder  = out_folder
        fig.savefig(Path(log_folder, 'log_{0}.pdf'.format(name)), bbox_inches = 'tight')
        ax.set_yscale('linear')
        ax.autoscale(True)
        if samples is not None:
            ax.set_xlim(xlim)
        fig.savefig(Path(plot_folder, '{0}.pdf'.format(name)), bbox_inches = 'tight')
        np.savetxt(Path(txt_folder, 'prob_{0}.txt'.format(name)), np.array([x, p[50], p[5], p[16], p[84], p[95]]).T, header = 'x 50 5 16 84 95')
    if show:
        ax.set_yscale('linear')
        ax.autoscale(True)
        if samples is not None:
            ax.set_xlim(xlim)
        plt.show()
    plt.close()
    
    # If selection function is available, plot reweighted distribution
    if selfunc is not None:
        for perc in percentiles:
            p[perc] = np.percentile((probs/f_x).T, perc, axis = 1)
        norm = p[50].sum()*dx
        for perc in percentiles:
            p[perc] = p[perc]/norm
        
        fig, ax = plt.subplots()
        ax.set_yscale('log')
        # CR
        ax.fill_between(x, p[95], p[5], color = 'mediumturquoise', alpha = 0.25)
        ax.fill_between(x, p[84], p[16], color = 'darkturquoise', alpha = 0.25)
        if injected is not None:
            # Injection
            ax.plot(x, p_x, lw = 0.5, color = 'red', label = injected_label)
        # Median
        ax.plot(x, p[50], lw = 0.7, color = 'steelblue', label = '${0}$'.format(median_label))
        if label is None:
            label = 'x'
        if unit is None or unit == '':
            ax.set_xlabel('${0}$'.format(label))
        else:
            ax.set_xlabel('${0}\ [{1}]$'.format(label, unit))
        ax.set_ylabel('$p({0})$'.format(label))
        ax.autoscale(True)
        ax.set_ylim(bottom = 1e-5, top = np.max(p[95])*1.1)
        ax.legend(loc = 0)
        if save:
            fig.savefig(Path(log_folder, 'log_true_{0}.pdf'.format(name)), bbox_inches = 'tight')
            ax.set_yscale('linear')
            ax.autoscale(True)
            fig.savefig(Path(plot_folder, 'true_{0}.pdf'.format(name)), bbox_inches = 'tight')
            np.savetxt(Path(txt_folder, 'prob_true_{0}.txt'.format(name)), np.array([x, p[50], p[5], p[16], p[84], p[95]]).T, header = 'x 50 5 16 84 95')
        if show:
            ax.set_yscale('linear')
            ax.autoscale(True)
            plt.show()
        plt.close()
    

def plot_multidim(draws, samples = None, bounds = None, out_folder = '.', name = 'density', labels = None, units = None, hierarchical = False, show = False, save = True, subfolder = False, n_pts = 200, true_value = None, figsize = 7, levels = [0.5, 0.68, 0.9], scatter_points = False, median_label = None):
    """
    Plot the recovered multidimensional distribution along with samples from the true distribution (if available) as corner plot.
    
    Arguments:
        :iterable draws:         container for mixture instances
        :int dim:                number of dimensions
        :np.ndarray samples:     samples from the true distribution (if available)
        :iterable bounds:        bounds for the recovered distribution. If None, bounds from mixture instances are used.
        :str or Path out_folder: output folder
        :str name:               name to be given to outputs
        :list-of-str labels:     LaTeX-style quantity label, for plotting purposes
        :list-of-str units:      LaTeX-style quantity unit, for plotting purposes
        :bool hierarchical:      hierarchical inference, for plotting purposes
        :bool save:              whether to save the plot or not
        :bool show:              whether to show the plot during the run or not
        :bool subfolder:         whether to save in a dedicated subfolder
        :int n_pts:              number of grid points (same for each dimension)
        :iterable true_value:    true value to plot
        :double figsize:         figure size (matplotlib)
        :iterable levels:        credible levels to plot
        :bool scatter_points:    scatter samples on 2d plots
        :str median_label:       label to assign to the reconstruction
    """
    
    dim = draws[0].dim
    if median_label is None:
        if hierarchical:
            median_label = '\mathrm{(H)DPGMM}'
        else:
            median_label = '\mathrm{DPGMM}'
    if labels is None:
        labels = ['$x_{0}$'.format(i+1) for i in range(dim)]
    else:
        labels = ['${0}$'.format(l) for l in labels]
    
    if units is not None:
        labels = [l[:-1]+'\ [{0}]$'.format(u) if not u == '' else l for l, u in zip(labels, units)]
    
    levels = np.atleast_1d(levels)

    all_bounds = np.atleast_2d([d.bounds for d in draws])
    x_min = np.min(all_bounds, axis = -1).max(axis = 0)
    x_max = np.max(all_bounds, axis = -1).min(axis = 0)
    
    probit = np.array([d.probit for d in draws]).any()
    
    if bounds is not None:
        bounds = np.atleast_2d(bounds)
        if bounds.shape == (1, 2):
            bounds = np.array([bounds[0] for _ in range(dim)])
        elif bounds.shape == (dim, 2):
            if not (bounds[:,0] >= x_min).all() and probit:
                warnings.warn("The provided lower bound is invalid for at least one draw. Default values will be used instead.")
            x_min[np.where(bounds[:,0] >= x_min)] = bounds[:,0][np.where(bounds[:,0] >= x_min)]
            if not (bounds[:,1] <= x_max).all() and probit:
                warnings.warn("The provided upper bound is invalid for at least one draw. Default values will be used instead.")
            x_max[np.where(bounds[:,1] <= x_max)] = bounds[:,1][np.where(bounds[:,1] <= x_max)]
        else:
            warnings.warn("The provided bounds have an invalid shape {0}. Shape must be (1,2) or (dim, 2).\nDefault bounds will be used instead.".format(bounds.shape))
    bounds = np.array([x_min, x_max]).T
    
    K = dim
    factor = 2.0          # size of one side of one panel
    lbdim = 0.5 * factor  # size of left/bottom margin
    trdim = 0.2 * factor  # size of top/right margin
    whspace = 0.1         # w/hspace size
    plotdim = factor * dim + factor * (K - 1.0) * whspace
    dim_plt = lbdim + plotdim + trdim
    
    fig, axs = plt.subplots(K, K, figsize=(figsize, figsize))
    # Format the figure.
    lb = lbdim / dim_plt
    tr = (lbdim + plotdim) / dim_plt
    fig.subplots_adjust(left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace)
    # Samples (if available)
    if samples is not None:
        bins = [int(np.sqrt(len(samples[:, c]))) for c in range(dim)]
        corner(samples, color = '#1f77b4', fig = fig, hist_kwargs = {'density': True, 'label':'$\mathrm{Samples}$', 'linewidth':0.7} , plot_density = False, contour_kwargs = {'linewidths':0.3, 'linestyles':'dashed'}, levels = [0.5,0.68,0.9], no_fill_contours = True, bins = bins)
        
    # 1D plots (diagonal)
    for column in range(K):
        ax = axs[column, column]
        # Marginalise over all uninterested columns
        dims = list(np.arange(dim))
        dims.remove(column)
        marg_draws = marginalise(draws, dims)
        # Credible regions
        lim = bounds[column]
        if samples is not None:
            lim_l = ax.get_xlim()
            lim[0] = np.max((lim[0], lim_l[0]))
            lim[1] = np.min((lim[1], lim_l[1]))
            
        x   = np.linspace(lim[0], lim[1], n_pts+2)[1:-1]
        dx  = x[1]-x[0]
        
        probs = np.array([d.pdf(x) for d in marg_draws])
        
        percentiles = [50, 5, 16, 84, 95]
        p = {}
        for perc in percentiles:
            p[perc] = np.percentile(probs, perc, axis = 0)
        norm = p[50].sum()*dx
        for perc in percentiles:
            p[perc] = p[perc]/norm
            
        # CR
        ax.fill_between(x, p[95], p[5], color = 'mediumturquoise', alpha = 0.25)
        ax.fill_between(x, p[84], p[16], color = 'darkturquoise', alpha = 0.25)
        if true_value is not None:
            if true_value[column] is not None:
                ax.axvline(true_value[column], c = 'orangered', lw = 0.5)
        ax.plot(x, p[50], lw = 0.7, color = 'steelblue', label = '${0}$'.format(median_label))
        ax.set_yticks([])
        if column == K - 1:
            if labels is not None:
                ax.set_xlabel(labels[-1])
        ticks = np.linspace(lim[0], lim[1], 5)
        ax.set_xticks(ticks)
        [l.set_rotation(45) for l in ax.get_xticklabels()]
        ax.set_xlim(lim[0], lim[1])
        if column < K-1:
            ax.set_xticklabels([])
            ax.set_yticklabels([])

    # 2D plots (off-diagonal)
    for row in range(K):
        for column in range(K):
            ax = axs[row,column]
            if column > row:
                ax.set_frame_on(False)
                ax.set_xticks([])
                ax.set_yticks([])
                continue
            elif column == row:
                continue
            
            # Marginalise
            dims = list(np.arange(dim))
            dims.remove(column)
            dims.remove(row)
            marg_draws = marginalise(draws, dims)
            
            # Credible regions
            lim = bounds[[row, column]]
            if samples is not None:
                lim_l = np.array([ax.get_ylim(), ax.get_xlim()])
                for i in range(2):
                    lim[i,0] = np.max((lim[i,0], lim_l[i,0]))
                    lim[i,1] = np.min((lim[i,1], lim_l[i,1]))
                
            grid, dgrid = recursive_grid(lim[::-1], np.ones(2, dtype = int)*int(n_pts))
            
            x = np.linspace(lim[0,0], lim[0,1], n_pts+2)[1:-1]
            y = np.linspace(lim[1,0], lim[1,1], n_pts+2)[1:-1]
            
            dd = np.array([d.pdf(grid) for d in marg_draws])
            median = np.percentile(dd, 50, axis = 0)
            median = median/(median.sum()*np.prod(dgrid))
            median = median.reshape(n_pts, n_pts)
            
            X,Y = np.meshgrid(x,y)
            with np.errstate(divide = 'ignore'):
                logmedian = np.nan_to_num(np.log(median), nan = -np.inf, neginf = -np.inf)
            _,_,levs = ConfidenceArea(logmedian, x, y, adLevels=levels)
            ax.contourf(Y, X, np.exp(logmedian), cmap = 'Blues', levels = 100)
            if true_value is not None:
                if true_value[row] is not None:
                    ax.axhline(true_value[row], c = 'orangered', lw = 0.5)
                if true_value[column] is not None:
                    ax.axvline(true_value[column], c = 'orangered', lw = 0.5)
                if true_value[column] is not None and true_value[row] is not None:
                    ax.plot(true_value[column], true_value[row], color = 'orangered', marker = 's', ms = 3)
            c1 = ax.contour(Y, X, logmedian, np.sort(levs), colors='steelblue', linewidths=0.3)
            if plot_settings.tex_flag:
                ax.clabel(c1, fmt = {l:'{0:.0f}\\%'.format(100*s) for l,s in zip(c1.levels, np.sort(levels)[::-1])}, fontsize = 3)
            else:
                ax.clabel(c1, fmt = {l:'{0:.0f}\%'.format(100*s) for l,s in zip(c1.levels, np.sort(levels)[::-1])}, fontsize = 3)
            # Samples (if available)
            if samples is not None and scatter_points:
                ax.scatter(samples[:,column], samples[:,row], marker = '+', c = 'orangered', linewidths = 1)
            # Ticks
            xticks = np.linspace(lim[1,0], lim[1,1], 5)
            ax.set_xticks(xticks)
            yticks = np.linspace(lim[0,0], lim[0,1], 5)
            ax.set_yticks(yticks)
            ax.set_xlim(*lim[1])
            ax.set_ylim(*lim[0])
            if column == 0:
                ax.set_ylabel(labels[row])
                [l.set_rotation(45) for l in ax.get_yticklabels()]
            if row == K - 1:
                [l.set_rotation(45) for l in ax.get_xticklabels()]
                ax.set_xlabel(labels[column])

    fig.axes[K-1].legend(*fig.axes[0].get_legend_handles_labels(), loc = 'center')
    fig.align_labels()
    
    if show:
        plt.show()
    if save:
        if not subfolder:
            fig.savefig(Path(out_folder, '{0}.pdf'.format(name)), bbox_inches = 'tight')
        else:
            if not Path(out_folder, 'density').exists():
                try:
                    Path(out_folder, 'density').mkdir()
                except FileExistsError:
                    pass
            fig.savefig(Path(out_folder, 'density', '{0}.pdf'.format(name)), bbox_inches = 'tight')
    plt.close()
    return fig
    
def plot_1d_dist(x, draws, injected = None, samples = None, out_folder = '.', name = 'density', label = None, unit = None, show = False, save = True, subfolder = False, true_value = None, true_value_label = '\mathrm{True\ value}', injected_label = '\mathrm{Simulated}', median_label = '\mathrm{Median}', logx = False, logy = False):
    """
    Plot a 1D distribution along with samples from the true distribution (if available).
    Differently from plot_median_cr, this method requires the distribution to be already evaluated.
    For FIGARO mixture instances, please use plot_median_cr.

    Arguments:
        :iterable x:                      values at which realisations are evaluated
        :iterable draws:                  container for realisations
        :callable or np.ndarray injected: injected distribution (if available)
        :np.ndarray samples:              samples from the true distribution (if available)
        :str or Path out_folder:          output folder
        :str name:                        name to be given to outputs
        :str label:                       LaTeX-style quantity label, for plotting purposes
        :str unit:                        LaTeX-style quantity unit, for plotting purposes
        :bool save:                       whether to save the plots or not
        :bool show:                       whether to show the plots during the run or not
        :bool subfolder:                  whether to save the plots in different subfolders (for multiple events)
        :float true_value:                true value to infer
        :str true_value_label:            label to assign to the true value marker
        :str injected_label:              label to assign to the injected distribution
        :str median_label:                label to assign to the median distribution
        :bool logx:                       x log scale
        :bool logy:                       y log scale
    """
    
    if not np.shape(x)[0] == np.shape(draws)[-1]:
        raise ValueError("x and each draw must have the same length")
    
    percentiles = [50, 5, 16, 84, 95]
    p = {}
    for perc in percentiles:
        p[perc] = np.percentile(draws, perc, axis = 0)
    
    fig, ax = plt.subplots()
    
    # Samples (if available)
    if samples is not None:
        ax.hist(samples, bins = int(np.sqrt(len(samples))), histtype = 'step', density = True, label = '$\mathrm{Samples}$')
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
    
    # CR
    ax.fill_between(x, p[95], p[5], color = 'mediumturquoise', alpha = 0.25)
    ax.fill_between(x, p[84], p[16], color = 'darkturquoise', alpha = 0.25)
    # Injection (if available)
    if injected is not None:
        if callable(injected):
            p_x = injected(x)
        else:
            p_x = injected
        if injected_label is not None:
            injected_label = '$'+injected_label+'$'
        ax.plot(x, p_x, lw = 0.5, color = 'red', label = injected_label)
        
    # Median
    if true_value is not None:
        if true_value_label is not None:
            true_value_label = '$'+true_value_label+'$'
        ax.axvline(true_value, ls = '--', color = 'r', lw = 0.5, label = true_value_label)
    if median_label is not None:
        median_label = '$'+median_label+'$'
    ax.plot(x, p[50], lw = 0.7, color = 'steelblue', label = median_label)
    if label is None:
        label = 'x'
    if unit is None or unit == '':
        ax.set_xlabel('${0}$'.format(label))
    else:
        ax.set_xlabel('${0}\ [{1}]$'.format(label, unit))
    ax.set_ylabel('$p({0})$'.format(label))
    if samples is not None:
        ax.set_xlim(xlim)
    ax.set_ylim(bottom = 1e-5, top = np.max(p[95])*1.1)
    ax.legend(loc = 0)
    if logy:
        ax.set_yscale('log')
    if logx:
        ax.set_xscale('log')
    
    fig.align_labels()
    
    if save:
        if subfolder:
            plot_folder = Path(out_folder, 'density')
            if not plot_folder.exists():
                try:
                    plot_folder.mkdir()
                except FileExistsError:
                    # Avoids issue with parallelisation
                    pass
            txt_folder = Path(out_folder, 'txt')
            if not txt_folder.exists():
                try:
                    txt_folder.mkdir()
                except FileExistsError:
                    pass
        else:
            plot_folder = out_folder
            txt_folder  = out_folder
        ax.autoscale(True)
        if samples is not None:
            ax.set_xlim(xlim)
        fig.savefig(Path(plot_folder, '{0}.pdf'.format(name)), bbox_inches = 'tight')
        np.savetxt(Path(txt_folder, 'prob_{0}.txt'.format(name)), np.array([x, p[50], p[5], p[16], p[84], p[95]]).T, header = 'x 50 5 16 84 95')
    if show:
        ax.autoscale(True)
        if samples is not None:
            ax.set_xlim(xlim)
        plt.show()
    plt.close()
    

def plot_n_clusters_alpha(n_cl, alpha, out_folder = '.', name = 'event', show = False, save = True):
    """
    Plot the number of clusters and the concentration parameter as functions of the number of samples.
    
    Arguments:
        :np.ndarray n_cl:        number of active clusters
        :np.ndarray alpha:       concentration parameter
        :str or Path out_folder: output folder
        :str name:               name to be given to outputs
        :bool save:              whether to save the plot or not
        :bool show:              whether to show the plot during the run or not
    """
    fig, ax = plt.subplots()
    ax1 = ax.twinx()
    ax.plot(np.arange(1, len(n_cl)+1), n_cl, ls = '--', marker = '', lw = 0.7, color = 'k')
    ax1.plot(np.arange(1, len(alpha)+1), alpha, ls = '--', marker = '', lw = 0.7, color = 'r')
    ax.set_xlabel('$t$')
    ax.set_ylabel('$N_{\mathrm{cl}}(t)$', color = 'k')
    ax1.set_ylabel('$\alpha(t)$', color = 'r')
    if show:
        plt.show()
    if save:
        fig.savefig(Path(out_folder, '{0}_n_cl_alpha.pdf'.format(name)), bbox_inches = 'tight')
    plt.close()

def pp_plot_cdf(draws, injection, n_points = 1000, out_folder = '.', name = 'event', show = False, save = True):
    """
    Make pp-plot comparing draws cdfs and injection cdf
    
    Arguments:
        :iterable draws:         container of mixture instances
        :callable injection:     injected density
        :int n_points:           number of points for linspace
        :str or Path out_folder: output folder
        :str name:               name to be given to outputs
        :bool save:              whether to save the plot or not
        :bool show:              whether to show the plot during the run or not
    """
    all_bounds = np.atleast_2d([d.bounds[0] for d in draws])
    x_min = np.max(all_bounds[:,0])
    x_max = np.min(all_bounds[:,1])
    x = np.linspace(x_min, x_max, n_points+2)[1:-1]
    
    functions     = np.array([mix(x) for mix in draws])
    median        = np.percentile(functions, 50, axis = 0)
    cdf_draws     = np.array([fast_cumulative(d) for d in functions])
    cdf_median    = fast_cumulative(median)
    cdf_injection = fast_cumulative(injection(x))
    
    fig = plt.figure()
    ax  = fig.add_subplot(111, projection = 'pp_plot')
    ax.add_confidence_band(len(cdf_median), cl=0.9, color = 'ghostwhite')
    ax.add_diagonal()
    for cdf in cdf_draws:
        ax.plot(cdf_injection, cdf, lw = 0.5, alpha = 0.5, color = 'darkturquoise')
    ax.plot(cdf_injection, cdf, color = 'steelblue', lw = 0.7)
    ax.set_xlabel('$\mathrm{Injected}$')
    ax.set_ylabel('$\mathrm{FIGARO}$')
    if show:
        plt.show()
    if save:
        fig.savefig(Path(out_folder, '{0}_ppplot.pdf'.format(name)), bbox_inches = 'tight')
    plt.close()

def pp_plot_levels(CR_levels, median_CR = None, out_folder = '.', name = 'MDC', show = False, save = True):
    """
    Make pp-plot.
    
    Arguments:
        :iterable CR:            2D array with credible levels for each event
        :iterable median_CR:     credible levels of medians
        :str or Path out_folder: output folder
        :str name:               name to be given to outputs
        :bool save:              whether to save the plot or not
        :bool show:              whether to show the plot during the run or not
    """
    CR_levels = np.atleast_1d(CR_levels)
    if len(CR_levels.shape) > 1:
        CR_levels = CR_levels.T
    n_evs     = CR_levels.shape[-1]
    L         = np.linspace(0,1,n_evs+2)
    
    fig = plt.figure()
    ax  = fig.add_subplot(111, projection = 'pp_plot')
    ax.add_confidence_band(n_evs, zorder = n_evs)
    ax.add_diagonal(zorder = n_evs+1)
    if len(CR_levels.shape) > 1:
        sorted = []
        for cr in CR_levels:
            if median_CR is not None:
                lw = 0.3
                c  = 'lightsteelblue'
            else:
                lw = 0.6
                c  = 'steelblue'
            x = np.append(0, np.append(cr, 1))
            ax.plot(np.sort(x), L, lw = lw, alpha = 0.5, color = c)
        if median_CR is not None:
            x = np.append(0, np.append(median_CR, 1))
            ax.plot(np.sort(x), L, lw = 0.8, color = 'steelblue', label = '$\mathrm{Median}$', zorder = n_evs+2)
        # Add label for draws
        handles, labels = ax.get_legend_handles_labels()
        line = Line2D([0], [0], label='$\mathrm{Draws}$', lw = lw, color = c)
        handles.extend([line])
        ax.legend(handles=handles, loc = 0, frameon = False)
    else:
        x = np.append(0, np.append(CR_levels, 1))
        ax.plot(np.sort(x), L, lw = 0.8, color = 'steelblue', zorder = n_evs+2)
    # Maquillage
    ax.set_xlabel('$P$')
    ax.set_ylabel('$\mathrm{Fraction\ of\ events\ within\ }CR_P$')
    if show:
        plt.show()
    if save:
        fig.savefig(Path(out_folder, '{0}_ppplot.pdf'.format(name)), bbox_inches = 'tight')
    plt.close()



import numpy as np

import optparse as op
import json
import dill
import importlib

from pathlib import Path
from tqdm import tqdm

from figaro.mixture import DPGMM
from figaro.transform import transform_to_probit
from figaro.utils import save_options, load_options, recursive_grid, get_priors
from figaro.plot import plot_median_cr, plot_multidim, pp_plot_levels
from figaro.credible_regions import FindLevelForHeight, FindNearest_Grid
from figaro.load import load_data, save_density, load_density

def main():

    parser = op.OptionParser()
    # Input/output
    parser.add_option("-i", "--input", type = "string", dest = "samples_folder", help = "Folder with single-event samples files. .txt files only", default = None)
    parser.add_option("-b", "--bounds", type = "string", dest = "bounds", help = "Density bounds. Must be a string formatted as '[[xmin, xmax], [ymin, ymax],...]'. For 1D distributions use '[xmin, xmax]'. Quotation marks are required and scientific notation is accepted", default = None)
    parser.add_option("--true_vals", type = "string", dest = "true_vals", help = "JSON file storing a dictionary with the injected values. Dictionary keys must match single-event samples files names", default = None)
    parser.add_option("-o", "--output", type = "string", dest = "output", help = "Output folder. Default: same directory as samples folder", default = None)
    parser.add_option("-j", dest = "json", action = 'store_true', help = "Save mixtures in json file", default = False)
    # Plot
    parser.add_option("--name", type = "string", dest = "name", help = "Name to be given to pp-plot files. Default: MDC", default = 'MDC')
    parser.add_option("-p", "--postprocess", dest = "postprocess", action = 'store_true', help = "Postprocessing", default = False)
    parser.add_option("-s", "--save_plot", dest = "save_plots", action = 'store_true', help = "Save single event plots", default = False)
    parser.add_option("--symbol", type = "string", dest = "symbol", help = "LaTeX-style quantity symbol, for plotting purposes", default = None)
    parser.add_option("--unit", type = "string", dest = "unit", help = "LaTeX-style quantity unit, for plotting purposes", default = None)
    # Settings
    parser.add_option("--draws", type = "int", dest = "n_draws", help = "Number of draws for each single-event distribution", default = 100)
    parser.add_option("--n_samples_dsp", type = "int", dest = "n_samples_dsp", help = "Number of samples to analyse (downsampling). Default: all", default = -1)
    parser.add_option("--exclude_points", dest = "exclude_points", action = 'store_true', help = "Exclude points outside bounds from analysis", default = False)
    parser.add_option("--grid_points", dest = "grid_points", type = "string", help = "Number of grid points for each dimension. Single integer or one int per dimension", default = None)
    parser.add_option("-e", "--events", dest = "run_events", action = 'store_false', help = "Skip single-event analysis", default = True)
    parser.add_option("--no_probit", dest = "probit", action = 'store_false', help = "Disable probit transformation", default = True)
    parser.add_option("--config", dest = "config", type = "string", help = "Config file. Warning: command line options are ignored if provided", default = None)

    (options, args) = parser.parse_args()

    # Paths
    if options.samples_folder is not None:
        options.samples_folder = Path(options.samples_folder).resolve()
    elif options.config is not None:
        options.samples_folder = Path('.').resolve()
    else:
        raise Exception("Please provide path to samples.")
    if options.output is not None:
        options.output = Path(options.output).resolve()
        if not options.output.exists():
            options.output.mkdir(parents=True)
    else:
        options.output = options.samples_folder.parent
    if options.config is not None:
        options.config = Path(options.config).resolve()
    output_plots = Path(options.output, 'plots')
    if not output_plots.exists():
        output_plots.mkdir()
    output_draws = Path(options.output, 'draws')
    if not output_draws.exists():
        output_draws.mkdir()

    # File extension
    if options.json:
        options.ext = 'json'
    else:
        options.ext = 'pkl'
    # Check that all files are .txt
    if not (np.array([f.suffix for f in options.samples_folder.glob('*')]) == '.txt').all():
        raise Exception("Only .txt files are currently supported for PP-plot analysis")

    if options.config is not None:
        load_options(options, options.config)
    save_options(options, options.output)

    # Read bounds
    if options.bounds is not None:
        options.bounds = np.array(np.atleast_2d(eval(options.bounds)), dtype = np.float64)
    elif options.bounds is None and not options.postprocess:
        raise Exception("Please provide bounds for the inference (use -b '[[xmin,xmax],[ymin,ymax],...]')")
    else:
        raise Exception("Please provide JSON file with true values")

    # Load samples
    events, names = load_data(options.samples_folder, n_samples = options.n_samples_dsp)
    try:
        dim = np.shape(events[0][0])[-1]
    except IndexError:
        dim = 1
    if dim > 3:
        raise Exception("PP-plots can be computed up to 3 dimensions")
    # Check all events have an entry in true_vals dict
    if not np.array([name in true_vals.keys() for name in names]).all():
        raise Exception("Please provide a dictionary storing all the true values. Dict keys must match event names. The following events appear not to have a true value:\n{0}".format(np.array(names)[np.where([name not in true_vals.keys() for name in names])]))

    if options.exclude_points:
        print("Ignoring points outside bounds.")
        for i, ev in enumerate(events):
            events[i] = ev[np.where((np.prod(options.bounds[:,0] < ev, axis = 1) & np.prod(ev < options.bounds[:,1], axis = 1)))]
        all_samples = np.atleast_2d(np.concatenate(events))
    else:
        # Check if all samples are within bounds
        all_samples = np.atleast_2d(np.concatenate(events))
        if options.probit:
            if not np.alltrue([(all_samples[:,i] > options.bounds[i,0]).all() and (all_samples[:,i] < options.bounds[i,1]).all() for i in range(dim)]):
                raise ValueError("One or more samples are outside the given bounds.")
    # Plot labels
    if dim > 1:
        if options.symbol is not None:
            symbols = options.symbol.split(',')
        else:
            symbols = options.symbol
        if options.unit is not None:
            units = options.unit.split(',')
        else:
            units = options.unit
    # Read grid points
    if options.grid_points is not None:
        pts = options.grid_points.split('.')
        if len(pts) == dim:
            options.grid_points = np.array([int(d) for d in pts])
        elif len(pts) == 1:
            options.grid_points = np.ones(dim, dtype = int)*int(pts[0])
        else:
            print("Wrong number of grid point provided. Falling back to default number")
            options.grid_points = np.ones(dim, dtype = int)*int((1000/dim**2)**dim)
    else:
        options.grid_points = np.ones(dim, dtype = int)*int((1000/dim**2)**dim)

    # Load true values
    if options.true_vals is not None:
        options.true_vals = Path(options.true_vals).resolve()
        with open(options.true_vals, 'r') as f:
            true_vals = json.load(f)

    # Reconstruction
    if not options.postprocess:
        mix        = DPGMM(options.bounds, probit = options.probit)
        grid, diff = recursive_grid(options.bounds, options.grid_points)
        logdiff    = np.sum(np.log(diff))
        posteriors = []
        CR_levels  = []
        CR_medians = []
        # Run single-event analysis
        for i in tqdm(range(len(events)), desc = 'Events'):
            ev   = events[i]
            name = names[i]
            # Load true value
            t = true_vals[name]
            if options.run_events:
                # Estimate prior pars from samples
                mix.initialise(prior_pars = get_priors(mix.bounds, samples = ev, probit = options.probit, hierarchical = False))
                # Draw samples
                draws = [mix.density_from_samples(ev) for _ in range(options.n_draws)]
                posteriors.append(draws)
                if options.save_plots:
                    if dim == 1:
                        plot_median_cr(draws,
                                       samples    = ev,
                                       out_folder = output_plots,
                                       name       = name,
                                       label      = options.symbol,
                                       unit       = options.unit,
                                       subfolder  = True,
                                       true_value = t,
                                       )
                    else:
                        plot_multidim(draws,
                                      samples    = ev,
                                      out_folder = output_plots,
                                      name       = name,
                                      labels     = symbols,
                                      units      = units,
                                      true_value = t,
                                      )
                # Save single-event draws
                save_density(draws, folder = output_draws, name = 'draws_'+name, ext = options.ext)
            else:
                draws = load_density(Path(output_draws, 'draws_'+name+'.'+options.ext))
            # Evaluate mixtures
            logP      = np.array([d.logpdf(grid) for d in draws])
            # Find true_value
            true_idx  = FindNearest_Grid(grid, t)
            logP_true = np.array([logP_i[true_idx] for logP_i in logP])
            # Compute credible levels
            CR = np.array([FindLevelForHeight(logP_i, logP_i_true, logdiff) for logP_i, logP_i_true in zip(logP, logP_true)])
            CR_medians.append(np.median(CR))
            CR_levels.append(CR)
            
        # Save all single-event draws together (might be useful for future hierarchical analysis)
        posteriors = np.array(posteriors)
        with open(Path(output_draws, 'posteriors_single_event.pkl'), 'wb') as f:
            dill.dump(posteriors, f)
        # Save credible levels
        CR_levels  = np.array(CR_levels)
        CR_medians = np.array(CR_medians)
        np.savetxt(Path(options.output, 'CR_levels.txt'), CR_levels, header = 'CR levels for events\nEach row is a different event, columns represent different draws')
        np.savetxt(Path(options.output, 'CR_medians.txt'), CR_medians)
    else:
        CR_levels = np.genfromtxt(Path(options.output, 'CR_levels.txt'))
        if len(CR_levels.shape) == 1:
            CR_levels = np.atleast_2d(CR_levels).T
        CR_medians = np.genfromtxt(Path(options.output, 'CR_medians.txt'))

    pp_plot_levels(CR_levels, median_CR = CR_medians, out_folder = options.output, name = options.name)

if __name__ == '__main__':
    main()
